"""
Abstractions for nodes representation.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from itertools import chain, count, cycle
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Union,
    cast,
)

import netunicorn.base

from .architecture import Architecture
from .environment_definitions import _available_environment_definitions
from .types import NodeProperty, NodeRepresentation, NodesRepresentation


class Node:
    """
    Represents a single node in a pool of nodes.

    :param name: name of the node
    :param properties: custom properties of the node
    :param architecture: node architecture
    """

    def __init__(
        self,
        name: str,
        properties: Dict[str, NodeProperty],
        architecture: Architecture = Architecture.UNKNOWN,
    ):
        self.name: str = name
        """
        Node name.
        """

        self.properties: Dict[str, NodeProperty] = properties
        """
        Node properties. Could be used to store custom information about the node.
        """

        self.additional_properties: Dict[str, NodeProperty] = {}
        """
        Additional node properties, often used for internal purposes and not to be exposed to the user.
        """

        self.architecture: Architecture = architecture
        """
        Node architecture.
        """

        self.available_environments: Set[type] = self._infer_environments()
        """
        Supported environments for the node (see :py:mod:`netunicorn.base.environment_definitions`).
        """

    def _infer_environments(self) -> Set[type]:
        result = set()
        # noinspection PyBroadException
        try:
            environments = self.properties.get(
                "netunicorn-environments", _available_environment_definitions.keys()
            )
            if hasattr(environments, "__iter__"):
                for environment_name in cast(Iterable[str], environments):
                    if environment_name in _available_environment_definitions:
                        result.add(_available_environment_definitions[environment_name])
        except Exception:
            return set(_available_environment_definitions.values())
        return result

    def __getitem__(self, item: str) -> NodeProperty:
        """
        Returns a node property by name.

        :param item: name of the property
        :return: property value
        """
        return self.properties.get(item, None)

    def __iter__(self) -> Iterator[NodeProperty]:
        raise TypeError("Node is not iterable")

    def __setitem__(self, key: str, value: NodeProperty) -> None:
        """
        Sets a node property.

        :param key: name of the property
        :param value: property value
        """
        self.properties[key] = value

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        """
        Checks if two nodes are equal.

        :param other: other node
        :return: True if the nodes have all parameters equal, False otherwise
        """
        return (
            isinstance(other, Node)
            and self.name == other.name
            and self.properties == other.properties
            and self.additional_properties == other.additional_properties
            and self.architecture == other.architecture
        )

    def __json__(self) -> NodeRepresentation:
        return {
            "name": self.name,
            "properties": self.properties,
            "additional_properties": self.additional_properties,
            "architecture": self.architecture.value,
        }

    @classmethod
    def from_json(cls, data: NodeRepresentation) -> Node:
        """
        Returns an instance of the object from a JSON representation.

        :param data: JSON representation of the object
        :return: instance of the object
        """
        instance = cls(
            data["name"], data["properties"], Architecture(data["architecture"])
        )
        instance.additional_properties = data["additional_properties"]
        return instance


class Nodes(ABC):
    """
    A base class that represents a pool of nodes. Not to be used directly, but to be inherited from.
    """

    @abstractmethod
    def __json__(self) -> NodesRepresentation:
        """
        Returns a JSON representation of the object.
        """
        pass

    @staticmethod
    def dispatch_and_deserialize(data: NodesRepresentation) -> Nodes:
        """
        Deserializes a JSON representation of the object and returns an instance of the object.

        :param data: JSON representation of the object
        :return: instance of the object
        """

        cls: Nodes = getattr(netunicorn.base.nodes, data["node_pool_type"])
        return cls.from_json(
            data["node_pool_data"], data.get("node_pool_metadata", None)
        )

    @classmethod
    @abstractmethod
    def from_json(
        cls,
        data: List[Union[NodeRepresentation, NodesRepresentation]],
        metadata: Optional[Any] = None,
    ) -> Nodes:
        """
        Class-specific implementation of deserialization from JSON.

        :param data: JSON representation of the object
        :param metadata: optional additional information for the pool instantiation
        :return: instance of the object
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, item: int) -> Union[Node, Nodes]:
        pass

    @abstractmethod
    def filter(self, function: Callable[[Node], bool]) -> Nodes:
        """
        Returns a pool of nodes that match the given filter function.

        :param function: filter function returning True if the node should be included in the result
        :return: pool of nodes
        """
        pass

    @abstractmethod
    def take(self, count: int) -> Sequence[Node]:
        """
        Returns a sequence of nodes consisting of the first `count` nodes.

        :param count: number of nodes to take
        :return: sequence of nodes
        """
        pass

    @abstractmethod
    def skip(self, count: int) -> Nodes:
        """
        Returns a pool of nodes consisting of the nodes after the first 'count' nodes.

        :param count: number of nodes to skip
        :return: pool of nodes
        """
        pass

    @abstractmethod
    def set_property(self, name: str, value: NodeProperty) -> Nodes:
        """
        Sets a property for all nodes in the pool.

        :param name: name of the property
        :param value: property value
        :return: self
        """
        pass


class CountableNodePool(Nodes):
    """
    Represents a pool of nodes that contains a fixed number of nodes.

    :param nodes: list of nodes
    """

    def __init__(self, nodes: List[Union[Node, Nodes]]):
        self.nodes = nodes
        """
        Nodes in the pool.
        """

    def __json__(self) -> NodesRepresentation:
        return {
            "node_pool_type": self.__class__.__name__,
            "node_pool_data": [x.__json__() for x in self.nodes],
            "node_pool_metadata": None,
        }

    @classmethod
    def from_json(
        cls,
        data: List[Union[NodeRepresentation, NodesRepresentation]],
        metadata: Optional[Any] = None,
    ) -> CountableNodePool:
        """
        Returns an instance of the object from a JSON representation.

        :param data: JSON representation of the object
        :param metadata: optional additional information for the pool instantiation
        :return: instance of the object
        """
        nodes: List[Union[Node, Nodes]] = []
        for element in data:
            if "node_pool_type" in element:
                # we know this is a NodesRepresentation
                nodes_representation_element = cast(NodesRepresentation, element)
                nodes.append(
                    Nodes.dispatch_and_deserialize(nodes_representation_element)
                )
            else:
                nodes.append(Node.from_json(element))
        return cls(nodes)

    def __str__(self) -> str:
        return str(self.nodes)

    def __len__(self) -> int:
        return len(self.nodes)

    def __getitem__(self, key: int) -> Union[Node, Nodes]:
        return self.nodes[key]

    def __setitem__(
        self, key: int, value: Union[Node, CountableNodePool, UncountableNodePool]
    ) -> None:
        self.nodes[key] = value

    def pop(self, index: int) -> Union[Node, Nodes]:
        """
        Removes and returns the node at the given index.

        :param index: index of the node to remove
        :return: popped node
        """
        return self.nodes.pop(index)

    def __repr__(self) -> str:
        return str(self.nodes)

    def filter(self, function: Callable[[Node], bool]) -> CountableNodePool:
        nodes: List[Union[Node, Nodes]] = []
        for node in self.nodes:
            if isinstance(node, Node):
                if function(node):
                    nodes.append(node)
            else:
                filtered_nodes = node.filter(function)
                if len(filtered_nodes) > 0:
                    nodes.append(filtered_nodes)
        return CountableNodePool(nodes)

    def take(self, count: int) -> Sequence[Node]:
        iterator: Iterator[Node] = chain.from_iterable(
            cast(
                Iterable[Iterable[Node]],
                ([x] if isinstance(x, Node) else x for x in self.nodes),
            )
        )
        nodes = []
        for _ in range(count):
            try:
                nodes.append(next(iterator))
            except StopIteration:
                break
        return nodes

    def skip(self, count: int) -> CountableNodePool:
        if count > len(self.nodes):
            print(
                f"Warning: asked to skip {count} nodes, but pool length is {len(self.nodes)}. "
                f"Returning empty pool."
            )
            return CountableNodePool([])
        return CountableNodePool(self.nodes[count:])

    def set_property(self, name: str, value: NodeProperty) -> CountableNodePool:
        for node in self.nodes:
            if isinstance(node, Node):
                node.properties[name] = value
            else:
                node.set_property(name, value)
        return self


class UncountableNodePool(Nodes):
    """
    | Represents a pool of nodes that is not countable (i.e., contains possibly infinite amount of nodes, e.g. cloud provider).
    | In the current implementation cannot have Nodes as elements.

    :param node_template: list of nodes that will be used as a template for generating new nodes
    """

    def __init__(self, node_template: List[Node], soft_limit: Optional[float] = None):
        self._node_template = node_template
        """
        Node template used for generating new nodes.
        """

        self._nodes = cycle(node_template)
        """
        An iterator over the node template.
        """

        self._counter = count(start=1, step=1)
        """
        Node name counter.
        """

        if soft_limit is not None and soft_limit <= 0:
            raise ValueError("Soft limit must be greater than 0.")
        self.soft_limit = float("inf") if soft_limit is None else soft_limit
        """
        Soft limit for nodes in the pool to prevent infinite loops or excessive node generation.
        """

    def __str__(self) -> str:
        return str(
            f"<Uncountable node pool with next node template: {self._node_template}>"
        )

    def __repr__(self) -> str:
        return str(self)

    def __iter__(self) -> Iterator[Node]:
        return self

    def __next__(self) -> Node:
        """
        Generate and returns the next node in the pool.

        :return: next node
        """
        node = deepcopy(next(self._nodes))
        node_number = next(self._counter)
        if node_number > self.soft_limit:
            raise StopIteration
        node.name += str(node_number)
        return node

    def __getitem__(self, key: int) -> Node:
        return self._node_template[key]

    def __setitem__(self, key: int, value: Node) -> None:
        self._node_template[key] = value

    def __len__(self) -> int:
        return len(self._node_template)

    def filter(self, function: Callable[[Node], bool]) -> UncountableNodePool:
        new_pool = UncountableNodePool([x for x in self._node_template if function(x)])
        new_pool._counter = self._counter  # keep the same counter
        new_pool.soft_limit = self.soft_limit
        return new_pool

    def take(self, count: int) -> Sequence[Node]:
        nodes = []
        for _ in range(count):
            nodes.append(next(self))
        return nodes

    def skip(self, count: int) -> UncountableNodePool:
        count = count % len(self._node_template)
        for _ in range(count):
            next(self._nodes)
        return self

    def __json__(self) -> NodesRepresentation:
        return {
            "node_pool_type": self.__class__.__name__,
            "node_pool_data": [x.__json__() for x in self._node_template],
            "node_pool_metadata": self.soft_limit,
        }

    @classmethod
    def from_json(
        cls,
        data: List[Union[NodeRepresentation, NodesRepresentation]],
        metadata: Optional[Any] = None,
    ) -> UncountableNodePool:
        for x in data:
            if "node_pool_type" in x:
                raise ValueError("UncountableNodePool cannot have Nodes as elements.")
        # now we have only NodeRepresentation in the list
        node_representation_data = cast(List[NodeRepresentation], data)
        if metadata is not None:
            metadata = float(metadata)
        return cls([Node.from_json(x) for x in node_representation_data], metadata)

    def set_property(self, name: str, value: NodeProperty) -> UncountableNodePool:
        for node in self._node_template:
            node.properties[name] = value
        return self



"""
Environment definitions to create an environment for a deployment.
"""

from __future__ import annotations

import platform
from dataclasses import dataclass, field
from typing import Dict, List, Optional, cast

from .types import (
    BuildContextRepresentation,
    DockerImageRepresentation,
    EnvironmentDefinitionRepresentation,
    RuntimeContextRepresentation,
    ShellExecutionRepresentation,
)

# TODO: make classes frozen to make hash stable


@dataclass
class RuntimeContext:
    """
    Stores a runtime context for a Deployment, such as port mapping or environment variables, that would be available during deployment

    :param ports_mapping: map of ports to be mapped to the host
    :param environment_variables: map of environment variables to be set
    :param additional_arguments: list of additional arguments to be passed to the runtime
    """

    ports_mapping: Dict[int, int] = field(default_factory=dict)
    """
    Desired port mapping (for all protocols)
    """

    environment_variables: Dict[str, str] = field(default_factory=dict)
    """
    Desired values of environment variables
    """

    network: Optional[str] = None
    """
    Desired network or interface name
    """

    additional_arguments: List[str] = field(default_factory=list)
    """
    Additional arguments (could be interpreted by runtimes)
    """

    def __json__(self) -> RuntimeContextRepresentation:
        return {
            "ports_mapping": self.ports_mapping,
            "environment_variables": self.environment_variables,
            "additional_arguments": self.additional_arguments,
            "network": self.network,
        }

    @classmethod
    def from_json(cls, data: RuntimeContextRepresentation) -> RuntimeContext:
        """
        Creates RuntimeContext from JSON representation.

        :param data: JSON representation of RuntimeContext
        :return: Deserialized RuntimeContext
        """

        ports_mapping = {int(k): v for k, v in data["ports_mapping"].items()}
        environment_variables = data["environment_variables"]
        additional_arguments = data["additional_arguments"]
        network = data.get("network")
        return cls(
            ports_mapping=ports_mapping,
            environment_variables=environment_variables,
            additional_arguments=additional_arguments,
            network=network,
        )


@dataclass
class EnvironmentDefinition:
    """
    A base class for all Environment Definitions.

    :param commands: commands that should be executed to create an environment
    """

    commands: List[str] = field(default_factory=list)
    """
    A list of commands that should be executed to create an environment.
    """

    runtime_context: RuntimeContext = field(default_factory=RuntimeContext)
    """
    Runtime context for this environment definition
    """

    def __hash__(self) -> int:
        return hash(tuple(self.commands))

    @classmethod
    def from_json(
        cls, data: EnvironmentDefinitionRepresentation
    ) -> EnvironmentDefinition:
        """
        Creates EnvironmentDefinition from JSON representation.

        :param data: JSON representation of EnvironmentDefinition
        :return: Deserialized EnvironmentDefinition
        """

        instance = cls.__new__(cls)
        instance.commands = data["commands"]
        return instance

    def __json__(self) -> EnvironmentDefinitionRepresentation:
        return {
            "commands": self.commands,
            "runtime_context": self.runtime_context.__json__(),
        }


@dataclass
class ShellExecution(EnvironmentDefinition):
    """
    This class represents Environment Definition that's created by executing commands in OS's shell.

    :param runtime_context: runtime context for this environment definition
    """

    @classmethod
    def from_json(cls, _data: EnvironmentDefinitionRepresentation) -> ShellExecution:
        """
        Creates ShellExecution from JSON representation.

        :param _data: JSON representation of ShellExecution
        :return: Deserialized ShellExecution
        """

        if "runtime_context" not in _data:
            raise ValueError("runtime_context is missing")
        data = cast(ShellExecutionRepresentation, _data)
        instance = cls.__new__(cls)
        instance.commands = data["commands"]
        instance.runtime_context = RuntimeContext.from_json(data["runtime_context"])
        return instance


@dataclass(frozen=True)
class BuildContext:
    """
    Stores a build context for a Deployment, such as Python version or cloudpickle version, used during deployment

    :param python_version: User's Python version
    :param cloudpickle_version: User's cloudpickle version
    """

    python_version: str = field(default_factory=platform.python_version)
    """
    User's Python version
    """

    cloudpickle_version: Optional[str] = field(
        default_factory=lambda: BuildContext._get_cloudpickle_version()
    )
    """
    User's cloudpickle version
    """

    @staticmethod
    def _get_cloudpickle_version() -> Optional[str]:
        try:
            import cloudpickle

            return cast(str, cloudpickle.__version__)
        except ImportError:
            return None

    def __json__(self) -> BuildContextRepresentation:
        return {
            "python_version": self.python_version,
            "cloudpickle_version": self.cloudpickle_version,
        }

    @classmethod
    def from_json(cls, data: BuildContextRepresentation) -> BuildContext:
        """
        Creates BuildContext from JSON representation.

        :param data: JSON representation of BuildContext
        :return: Deserialized BuildContext
        """

        python_version = data["python_version"]
        cloudpickle_version = data["cloudpickle_version"]
        return cls(
            python_version=python_version, cloudpickle_version=cloudpickle_version
        )


@dataclass
class DockerImage(EnvironmentDefinition):
    """
    | This class represents Environment Definition that is created by using a Docker image.
    | If image name is not provided, then it would be created automatically

    :param image: Docker image name
    :param build_context: build context for this environment definition
    :param runtime_context: runtime context for this environment definition
    """

    image: Optional[str] = None
    """
    Docker image name
    """

    build_context: BuildContext = field(default_factory=BuildContext)
    """
    Build context for this environment definition
    """

    runtime_context: RuntimeContext = field(default_factory=RuntimeContext)
    """
    Runtime context for this environment definition
    """

    def __hash__(self) -> int:
        if self.image:
            return hash(self.image)

        return hash(
            (
                self.build_context.python_version,
                self.build_context.cloudpickle_version,
                tuple(self.commands),
            )
        )

    def __json__(self) -> DockerImageRepresentation:
        return {
            "image": self.image,
            "build_context": self.build_context.__json__(),
            "runtime_context": self.runtime_context.__json__(),
            "commands": self.commands,
        }

    @classmethod
    def from_json(cls, _data: EnvironmentDefinitionRepresentation) -> DockerImage:
        """
        Creates DockerImage from JSON representation.

        :param _data: JSON representation of DockerImage
        :return: Deserialized DockerImage
        """

        if "image" not in _data:
            raise ValueError("image is missing")
        data = cast(DockerImageRepresentation, _data)
        instance = cls.__new__(cls)
        instance.commands = data["commands"]
        instance.image = data["image"]
        instance.build_context = BuildContext.from_json(data["build_context"])
        instance.runtime_context = RuntimeContext.from_json(data["runtime_context"])
        return instance


_available_environment_definitions = {
    ShellExecution.__name__: ShellExecution,
    DockerImage.__name__: DockerImage,
}



import sys
from abc import ABC, abstractmethod
from typing import Dict, List, Optional

from netunicorn.base.experiment import Experiment

if sys.version_info >= (3, 10):
    from typing import TypeAlias

    from typing_extensions import TypedDict
else:
    from typing_extensions import TypeAlias, TypedDict

ConnectorContext: TypeAlias = Optional[Dict[str, Dict[str, str]]]


class ExecutorsCancellationRequest(TypedDict):
    executors: List[str]
    cancellation_context: ConnectorContext


class BasePreprocessor(ABC):
    @abstractmethod
    def __call__(self, experiment_id: str, experiment: Experiment) -> Experiment:
        return experiment


class BasePostprocessor(ABC):
    @abstractmethod
    def __call__(self, experiment_id: str, experiment: Experiment) -> Experiment:
        return experiment



"""
Various auxiliary types.
"""

from __future__ import annotations

import sys
from enum import IntEnum
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import BaseModel
from returns.result import Result
from typing_extensions import TypedDict

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

NodeProperty = Union[str, float, int, Set[str], None]

TaskElementResult: TypeAlias = Result[Any, Any]
ExecutionGraphResult = Dict[str, List[TaskElementResult]]
PipelineResult = ExecutionGraphResult


class NodeRepresentation(TypedDict):
    """
    Node JSON representation.
    """

    name: str
    """
    Node name.
    """

    properties: Dict[str, NodeProperty]
    """
    Node properties.
    """

    additional_properties: Dict[str, NodeProperty]
    """
    Internal node properties.
    """

    architecture: str
    """
    Node architecture.
    """


class EnvironmentDefinitionRepresentation(TypedDict):
    """
    Environment definition JSON representation.
    """

    commands: List[str]
    """
    Environment definition commands.
    """

    runtime_context: RuntimeContextRepresentation
    """
    Runtime context JSON representation.
    """


class DeploymentRepresentation(TypedDict):
    """
    Deployment JSON representation.
    """

    node: NodeRepresentation
    """
    Node JSON representation.
    """

    prepared: bool
    """
    Prepared flag.
    """

    executor_id: str
    """
    Executor ID.
    """

    error: Optional[str]
    """
    Error message.
    """

    execution_graph: bytes
    """
    Serialized execution graph.
    """

    keep_alive_timeout_minutes: int
    """
    Keep alive timeout value.
    """

    cleanup: bool
    """
    Cleanup flag.
    """

    environment_definition: EnvironmentDefinitionRepresentation
    """
    Environment definition JSON representation.
    """

    environment_definition_type: str
    """
    Type of environment definition.
    """


class NodesRepresentation(TypedDict):
    """
    Nodes JSON representation.
    """

    node_pool_type: str
    """
    Type of node pool.
    """

    node_pool_data: List[Union[NodeRepresentation, NodesRepresentation]]
    """
    Node pool data (nodes)
    """

    node_pool_metadata: Optional[Any]
    """
    Optional additional information for the pool instantiation
    """


class BuildContextRepresentation(TypedDict):
    """
    Build context JSON representation.
    """

    python_version: str
    """
    Python version.
    """

    cloudpickle_version: Optional[str]
    """
    Cloudpickle version.
    """


class RuntimeContextRepresentation(TypedDict):
    """
    Runtime context JSON representation.
    """

    ports_mapping: Dict[int, int]
    """
    Ports mapping dictionary.
    """

    environment_variables: Dict[str, str]
    """
    Environment variables dictionary.
    """

    network: Optional[str]
    """
    Network interface name.
    """

    additional_arguments: List[str]
    """
    Additional arguments list.
    """


class ShellExecutionRepresentation(EnvironmentDefinitionRepresentation):
    """
    Shell execution Environment definition JSON representation.
    """

    pass


class DockerImageRepresentation(EnvironmentDefinitionRepresentation):
    """
    Docker image Environment definition JSON representation.
    """

    image: Optional[str]
    """
    Image name.
    """

    build_context: BuildContextRepresentation
    """
    Build context JSON representation.
    """


class ExperimentRepresentation(TypedDict):
    """
    Experiment JSON representation.
    """

    deployment_map: List[DeploymentRepresentation]
    """
    List of deployments of the experiment.
    """

    deployment_context: Optional[Dict[str, Dict[str, str]]]
    """
    Deployment context.
    """


class DeploymentExecutionResultRepresentation(TypedDict):
    """
    Deployment execution result JSON representation.
    """

    node: NodeRepresentation
    """
    Node JSON representation.
    """

    execution_graph: str
    """
    Serialized execution_graph.
    """

    result: Optional[str]
    """
    Result of execution.
    """

    error: Optional[str]
    """
    Error message.
    """


class ExperimentExecutionInformationRepresentation(TypedDict):
    """
    Experiment execution information JSON representation.
    """

    status: int
    """
    Status of execution value (see ExperimentExecutionStatus).
    """

    experiment: Optional[ExperimentRepresentation]
    """
    Experiment definition.
    """

    execution_result: Union[None, str, List[DeploymentExecutionResultRepresentation]]
    """
    Execution result.
    """


class ExecutorState(IntEnum):
    """
    Executor state.
    """

    LOOKING_FOR_EXECUTION_GRAPH = 0
    """
    Looking for the execution graph locally or downloading from netunicorn gateway.
    """

    EXECUTING = 1
    """
    Currently executing the graph.
    """

    REPORTING = 2
    """
    Execution is finished and executor is reporting results.
    """

    FINISHED = 3
    """
    Execution is finished and results are reported (if needed).
    """


class FlagValues(BaseModel):
    """
    | Flag values for experiment-wide flags. These flags could be get and set by the user or any executor.
    | One can use these flags to synchronize different executors, pass additional information,
        or manually control the experiment.
    |
    | Flag value could contain text, integer, or both. Integer values could be atomically incremented or decremented.
    """

    text_value: Optional[str] = None
    """
    Text value of the flag.
    """

    int_value: int = 0
    """
    Integer value of the flag.
    """



from __future__ import annotations

from abc import abstractmethod
from logging import Logger
from typing import Any, Optional, Protocol, Tuple

from netunicorn.base.deployment import Deployment
from netunicorn.base.nodes import Nodes
from returns.result import Result

from .types import StopExecutorRequest


class NetunicornConnectorProtocol(Protocol):
    """
    Represents netunicorn infrastructure connector protocol.
    All connectors must implement this protocol.
    Connectors provide available node pools of the infrastructure and
    ability to deploy and execute experiments.

    If connector raises an exception during the execution of any method,
    it will be removed from the list of available connectors.

    IMPORTANT: all methods of this class have *args and **kwargs to allow
    adding new parameters without breaking backward compatibility.
    """

    @abstractmethod
    def __init__(
        self,
        connector_name: str,
        configuration: str | None,
        netunicorn_gateway: str,
        logger: Optional[Logger] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Connector constructor.
        :param connector_name: system-wide unique connector name
        :param configuration: connector-specific configuration string (could be path to config file or anything)
        :param netunicorn_gateway: netunicorn gateway endpoint
        """
        _ = connector_name
        _ = configuration
        _ = netunicorn_gateway
        _ = logger

    @abstractmethod
    async def initialize(self, *args: Any, **kwargs: Any) -> None:
        """
        This method is guaranteed to be called immediately after the constructor
        to provide async initialization capabilities.
        If you don't know why you can need this method, leave it empty.
        """
        pass

    @abstractmethod
    async def health(self, *args: Any, **kwargs: Any) -> Tuple[bool, str]:
        """
        Health check of the connector.
        :return: True if connector is healthy, False otherwise, and a description of the health status
        """
        pass

    @abstractmethod
    async def shutdown(self, *args: Any, **kwargs: Any) -> None:
        """
        Shutdown the connector. Will be called in the end of the program.
        """
        pass

    @abstractmethod
    async def get_nodes(
        self,
        username: str,
        authentication_context: Optional[dict[str, str]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Nodes:
        """
        Get available nodes for the user.
        :param username: username
        :param authentication_context: authentication context provided by the user
        :return: Pool of nodes
        """
        pass

    @abstractmethod
    async def deploy(
        self,
        username: str,
        experiment_id: str,
        deployments: list[Deployment],
        deployment_context: Optional[dict[str, str]],
        authentication_context: Optional[dict[str, str]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Result[Optional[str], str]]:
        """
        This method deploys the given list of deployments to the infrastructure.
        See the documentation for available deployment environments for correct deployment implementation.

        :param username: username of the user who deploys the experiment
        :param experiment_id: ID of the experiment
        :param deployments: list of deployments to deploy
        :param deployment_context: optional deployment context provided directly from the user.
        Can be used to pass additional information to the connector, netunicorn does not modify or validate this field.
        You can define in the connector's documentation some additional deployment flags and ask user to set them in this field.
        E.g.: network configuration for virtual deployments (link capacity, etc.)
        :param authentication_context: optional authentication context provided by the user
        :return: dictionary of executor_id
        ((unique, parsed from deployment) -> Result[optional success message, error message])
        """
        pass

    @abstractmethod
    async def execute(
        self,
        username: str,
        experiment_id: str,
        deployments: list[Deployment],
        execution_context: Optional[dict[str, str]],
        authentication_context: Optional[dict[str, str]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Result[Optional[str], str]]:
        """
        This method starts execution of the given list of deployments on the infrastructure.
        Usually, that means that connector should run `python3 -m netunicorn.executor` on the node
         with the next environment variables:
        - NETUNICORN_GATEWAY_ENDPOINT: netunicorn gateway endpoint
        - NETUNICORN_EXECUTOR_ID: unique executor ID
        - NETUNICORN_EXPERIMENT_ID: experiment ID

        :param username: username of the user who deploys the experiment
        :param experiment_id: ID of the experiment
        :param deployments: list of deployments to start execution
        :param execution_context: optional execution context provided directly from the user.
        Can be used to pass additional information to the connector, netunicorn does not modify or validate this field.
        You can define in the connector's documentation some additional deployment flags and ask user to set them in this field.
        :param authentication_context: optional authentication context provided by the user
        :return: dictionary of (executor_id (unique, parsed from deployment) -> Result[optional success message, error message])
        """
        pass

    @abstractmethod
    async def stop_executors(
        self,
        username: str,
        requests_list: list[StopExecutorRequest],
        cancellation_context: Optional[dict[str, str]],
        authentication_context: Optional[dict[str, str]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Result[Optional[str], str]]:
        """
        This method stops execution of the given list of executors on the infrastructure.
        :param username: username of the user
        :param requests_list: list of StopExecutorRequests
        :param cancellation_context: optional cancellation context provided directly from the user.
        User can define arbitrary fields in this dictionary and connector can parse them.
        E.g.: a stopping reason for the experiment, or soft-kill vs hard-kill.
        :param authentication_context: optional authentication context provided by the user
        :return: dictionary of executor_id (unique, parsed from deployment) -> Result[optional success message, error message]
        """
        pass

    @abstractmethod
    async def cleanup(
        self,
        experiment_id: str,
        deployments: list[Deployment],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        This method is called after the experiment is finished.
        It is used to clean up the infrastructure after the experiment.
        E.g., delete Docker containers, images, etc.
        Experiment is always marked as cleaned up, even if this method fails,
        and this method should not return any result.

        :param experiment_id: ID of the experiment
        :param deployments: list of deployments to use for cleanup
        :return: None
        """
        pass



"""
Abstraction for an execution graph that contains tasks and their order.
"""

from __future__ import annotations

import uuid
from typing import Optional

import networkx as nx

from .environment_definitions import DockerImage, EnvironmentDefinition
from .task import Task, TaskDispatcher


class ExecutionGraph:
    """
    ExecutionGraph is a class that allows you to flexibly define a graph of tasks and their order.
    It has the next rules:

    | 1. Execution graph is a directed graph.
    | 2. Execution graph always starts with a node "root" (this node is automatically added in the new graph during initialization).
    | 3. Any node of type TaskDispather would be dispatched to Task. Any node of type Task would be executed. Any nodes of other types would not be executed but would be treated as synchronization points.
    | 4. Number of components in the graph must be 1, the graph should be weakly connected, and all nodes should be accessible from the root node.
    | 5. Any edge have either "strong" or "weak" type. If no type provided, it's a "strong" edge.
    Type is provided as an attribute "type" with value "strong" or "weak".
    Executor will not treat incoming "weak" edges as requirements for the execution of the task, but will traverse them to start executing next tasks.
    "Weak" edges are required for cycle dependencies to avoid deadlocks.
    You can consider weak links as links for defining the execution flow, and strong links to define both execution flow and prerequisites.

    | 6. Any edge can have attribute "counter" with integer value. This value would be used to determine how many times this edge should be traversed. If this attribute is not present, then edge would be traversed infinitely. You can use this attribute to implement finite loops in the graph.

    | 7. Any edge can have attribute "traverse_on" with either string value "success", "failure", or "any". This attribute controls whether executor will traverse this edge on success or failure of the task. If this attribute is not present, the executor will obey the usual rules considering the "early_stopping" parameter (trat all edges as "success" for "early_stopping=True" or "any" for "early_stopping=False"). If this attribute is present, then the executor will traverse this edge only if the task was executed successfully ("success"), failed ("failure"), or in any case ("any"). The next rules apply:
    If "early_stopping" is True, all edges are treated as having "traverse_on" attribute set to "success". Setting the "traverse_on" attribute to "failure" would make the executor to traverse this edge in case of the task failure and will not break the execution of the graph. Also, setting the "traverse_on" attribute to "any" would make executor to traverse this edge in case of the task failure or success and will not break the execution of the graph.
    If "early_stopping" is False, all edges are treated as having "traverse_on" attribute set to "any". Setting this attribute to "success" or "failure" would make the executor to traverse this edge only in case of the task success or failure respectively.

    :param early_stopping: If True, then the execution of the graph would be stopped if any task fails. If False, then the execution of the graph would be continued even if some tasks fail. In both cases, "traverse_on" attribute of the edges could be used to control the traversal of the graph.
    :param report_results: If True, then the executor would connect core services to report execution results in the end.
    :param environment_definition: environment definition for the execution graph
    """

    def __init__(
        self,
        early_stopping: bool = True,
        report_results: bool = True,
        environment_definition: Optional[EnvironmentDefinition] = None,
    ):
        self.name: str = str(uuid.uuid4())
        """
        Execution Graph name.
        """

        self.early_stopping: bool = early_stopping
        """
        Whether to stop executing tasks after a first failure.
        """

        self.report_results: bool = report_results
        """
        Whether executor should connect core services to report execution results in the end.
        """

        self.environment_definition: EnvironmentDefinition = (
            environment_definition or DockerImage()
        )
        """
        Environment definition for the execution graph.
        """

        self.graph = nx.DiGraph()
        """
        Graph of tasks and their order.
        """

        self.override_graph_validation = False
        """
        Disable graph validation. Executor and other components will not validate the graph before execution.
        """

        self.graph.add_node("root")

    @staticmethod
    def is_execution_graph_valid(obj: ExecutionGraph) -> bool:
        """
        Validates execution graph according to the ExecutionGraph rules.

        :return: True if execution graph is valid, raises an exception otherwise
        """

        if obj.override_graph_validation:
            return True

        if not isinstance(obj, ExecutionGraph):
            raise ValueError("The object is not an instance of ExecutionGraph")

        graph = obj.graph

        if not isinstance(graph, nx.DiGraph) or isinstance(graph, nx.MultiDiGraph):
            raise ValueError("Execution graph must be a directed graph (DiGraph)")

        if not graph.has_node("root"):
            raise ValueError("Execution graph must have a root node")

        if not nx.is_weakly_connected(graph):
            raise ValueError(
                "Execution graph must be a weakly connected directed graph"
            )

        successors = set(nx.dfs_postorder_nodes(graph, "root"))
        if diff := set(graph.nodes).difference(successors):
            raise ValueError(
                f"All tasks must be accessible from the root node. Inaccessible nodes: {diff}"
            )

        # if remove all weak links, the graph should be acyclic
        graph_copy = graph.copy()
        graph_copy.remove_edges_from(
            [
                (u, v)
                for u, v, d in graph_copy.edges(data=True)
                if d.get("type", "strong") == "weak"
            ]
        )
        if not nx.is_directed_acyclic_graph(graph_copy):
            raise ValueError(
                "Execution graph must be acyclic after removing all weak links. Otherwise your cycles will introduce deadlocks."
            )
        successors = set(nx.dfs_postorder_nodes(graph_copy, "root"))
        if diff := set(graph_copy.nodes).difference(successors):
            raise ValueError(
                f"After removing weak links, all nodes should be accessible from the root. Inaccessible nodes: {diff}"
            )

        # "counter" attribute on each edge should be positive number
        for u, v, d in graph.edges(data=True):
            if "counter" in d:
                try:
                    counter = int(d["counter"])
                except ValueError:
                    raise ValueError(
                        f"Edge ({u}, {v}) has attribute 'counter' with value '{d['counter']}'. "
                        f"This value should be a valid integer."
                    )

                if counter <= 0:
                    raise ValueError(
                        f"Edge ({u}, {v}) has attribute 'counter' with value '{d['counter']}'. "
                        f"This value should be a positive integer."
                    )

        # all edges having "traverse_on" attribute should have value "success", "failure", or "any"
        for u, v, d in graph.edges(data=True):
            if "traverse_on" in d:
                traverse_on = d["traverse_on"]
                if traverse_on not in ["success", "failure", "any"]:
                    raise ValueError(
                        f"Edge ({u}, {v}) has attribute 'traverse_on' with value '{traverse_on}'. This value should be either 'success', 'failure', or 'any'."
                    )

        # no edges starting at synchronization points (non-Task) points should have "traverse_on" attribute
        for u, v, d in graph.edges(data=True):
            if (
                not (isinstance(u, Task) or isinstance(u, TaskDispatcher))
                and "traverse_on" in d
            ):
                raise ValueError(
                    f"Edge ({u}, {v}) has attribute 'traverse_on' with value '{d['traverse_on']}'. This attribute should not be present for edges starting at synchronization points (non-Task nodes)."
                )

        return True

    def __copy__(self) -> ExecutionGraph:
        graph = ExecutionGraph(
            early_stopping=self.early_stopping,
            report_results=self.report_results,
            environment_definition=self.environment_definition,
        )
        graph.graph = self.graph.copy()
        return graph

    def __str__(self) -> str:
        return f"ExecutionGraph(name={self.name}, {self.graph.__str__()})"

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return self.__str__()

    # graph visualization
    @staticmethod
    def _get_edge_label_string(edge: dict) -> str:  # type: ignore[type-arg]
        values = [
            f"'{key}': {value}" for key, value in edge.items() if value is not None
        ]

        label = ", ".join(values)
        return f"{{{label}}}" if values else ""

    def draw(self, nx_layout_function=nx.shell_layout) -> None:  # type: ignore[no-untyped-def]
        """
        Draw execution graph using networkx library.

        :param nx_layout_function: networkx layout function to use for drawing (from networkx.drawing.layout)
        :return: None
        """

        node_labels = {
            task: task.name if hasattr(task, "name") else task
            for task in self.graph.nodes
        }

        edge_labels = {
            (u, v): self._get_edge_label_string(d)
            for u, v, d in self.graph.edges(data=True)
        }

        layout = nx_layout_function(self.graph)
        nx.draw(self.graph, layout)
        nx.draw_networkx_labels(self.graph, layout, labels=node_labels)
        nx.draw_networkx_edge_labels(self.graph, layout, edge_labels=edge_labels)

        return



"""
Utility functions and classes for netunicorn. Not to be directly used by users.
"""

import dataclasses
from base64 import b64encode
from json import JSONEncoder
from typing import Any, List

from netunicorn.base.environment_definitions import EnvironmentDefinition
from returns.result import Result

SerializedExecutionGraphType = bytes
LogType = List[str]


class UnicornEncoder(JSONEncoder):
    """
    Custom JSON encoder for netunicorn objects.
    """

    def default(self, obj: Any) -> Any:  # pylint: disable=E0202
        """
        Overriden default method for JSONEncoder.

        :param obj: Object to be serialized.
        :return: Serialized object.
        """
        if isinstance(obj, Exception):
            return str(obj.__reduce__())
        if isinstance(obj, set):
            return list(obj)
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)  # type: ignore  # https://github.com/python/mypy/issues/17550
        if hasattr(obj, "__json__"):
            return obj.__json__()
        if isinstance(obj, bytes):
            return b64encode(obj).decode("utf-8")
        if isinstance(obj, Result):
            # noinspection PyProtectedMember
            return {"result_type": obj.__class__.__name__, "result": obj._inner_value}
        if isinstance(obj, EnvironmentDefinition):
            return {
                "environment_definition_type": obj.__class__.__name__,
                "environment_definition": obj,
            }
        return JSONEncoder.default(self, obj)



"""
Single deployment of an execution graph on a node.
"""

from __future__ import annotations

import warnings
from base64 import b64decode
from copy import deepcopy
from typing import Optional

import netunicorn.base.environment_definitions
import networkx as nx

from .execution_graph import ExecutionGraph
from .nodes import Node
from .task import Task, TaskDispatcher
from .types import DeploymentRepresentation
from .utils import SerializedExecutionGraphType

try:
    import cloudpickle  # it's needed only on client side, but this module is also imported on engine side
    import netunicorn.library

    cloudpickle.register_pickle_by_value(netunicorn.library)
except ImportError:
    pass


class Deployment:
    """
    Single deployment of an execution graph on a node.

    :param node: Node to deploy execution graph on
    :param pipeline: pipeline (execution graph) to deploy
    :param keep_alive_timeout_minutes: time to wait for executor update before timeout
    :param cleanup: whether to remove artifacts (e.g., Docker image and containers) after execution
    """

    def __init__(
        self,
        node: Node,
        pipeline: ExecutionGraph,
        keep_alive_timeout_minutes: int = 10,
        cleanup: bool = True,
    ):
        self.node: Node = node
        """
        Node to deploy execution graph on
        """

        self.prepared: bool = False
        """
        if False, deployment is not prepared yet or failed during preparation
        """

        self.executor_id: str = ""
        """
        ID of executor on node
        """

        self.error: Optional[Exception] = None
        """
        if deployment failed, this field contains error
        """

        self.execution_graph: SerializedExecutionGraphType = b""
        """
        Serialized ExecutionGraph to be deployed
        """

        self.environment_definition: (
            netunicorn.base.environment_definitions.EnvironmentDefinition
        ) = deepcopy(pipeline.environment_definition)
        """
        Environment definition to use for deployment
        """

        self.keep_alive_timeout_minutes: int = keep_alive_timeout_minutes
        """
        time to wait for executor update before timeout
        """

        self.cleanup: bool = cleanup
        """
        if True, corresponding artifacts (Docker image and containers) will be removed after execution
        """

        self._validate_deployment(node, pipeline)

        execution_graph = deepcopy(pipeline)

        # dispatch tasks if needed
        task_dispatchers = {
            x: x.dispatch(node)
            for x in execution_graph.graph.nodes
            if isinstance(x, TaskDispatcher)
        }
        execution_graph.graph = nx.relabel_nodes(
            execution_graph.graph, task_dispatchers
        )

        # combine commands into environment definition
        for task in (x for x in execution_graph.graph.nodes if isinstance(x, Task)):
            self.environment_definition.commands.extend(task.requirements)

        self.execution_graph = cloudpickle.dumps(execution_graph)

    @staticmethod
    def _validate_deployment(node: Node, pipeline: ExecutionGraph) -> None:
        if type(pipeline.environment_definition) not in node.available_environments:
            raise ValueError(
                f"Node {node.name} does not support environment {type(pipeline.environment_definition).__name__}"
            )

        ExecutionGraph.is_execution_graph_valid(pipeline)

    def __str__(self) -> str:
        return f"Deployment: Node={self.node.name}, executor_id={self.executor_id}, prepared={self.prepared}, error={self.error}"

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> DeploymentRepresentation:
        return {
            "node": self.node.__json__(),
            "prepared": self.prepared,
            "executor_id": self.executor_id,
            "error": str(self.error) if self.error else None,
            "execution_graph": self.execution_graph,
            "keep_alive_timeout_minutes": self.keep_alive_timeout_minutes,
            "cleanup": self.cleanup,
            "environment_definition": self.environment_definition.__json__(),
            "environment_definition_type": self.environment_definition.__class__.__name__,
        }

    @classmethod
    def from_json(cls, data: DeploymentRepresentation) -> Deployment:
        """
        Create Deployment from JSON representation

        :param data: JSON representation of Deployment
        :return: Deserialized Deployment
        """
        instance = cls.__new__(cls)

        instance.node = Node.from_json(data["node"])
        instance.prepared = data["prepared"]
        instance.executor_id = data["executor_id"]
        instance.error = Exception(data["error"]) if data["error"] else None
        instance.keep_alive_timeout_minutes = data["keep_alive_timeout_minutes"]
        instance.cleanup = data.get("cleanup", True)
        instance.environment_definition = getattr(
            netunicorn.base.environment_definitions, data["environment_definition_type"]
        ).from_json(data["environment_definition"])

        try:
            instance.execution_graph = b64decode(data["execution_graph"])
        except KeyError:
            if "pipeline" in data:
                warnings.warn(
                    "The deployment was created with an older version of netunicorn. "
                    "Execution graph information would not be available."
                )
                instance.execution_graph = b""
            else:
                raise

        return instance



"""
Abstractions for tasks representation.
"""

from __future__ import annotations

import copy
import uuid
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Union

from returns.result import Failure, Success

from .nodes import Node
from .types import ExecutionGraphResult

# Keep classes for export
Success = Success
Failure = Failure


class Task(ABC):
    """
    | This is a base class for all tasks. All new task classes should inherit from this class.
    | The task instance should encapsulate all the logic and data needed to execute the task.
    | Task entrypoint is the run() method.
    | Task class can have requirements - commands to be executed to change environment to support this task.
    | These requirements would be executed with OS shell during environment setup.
    | Each task is to be implemented for a specific architecture, platform, or combination (like Linux + arm64).
    | TaskDispatcher can be used for selecting a specific task for the given architecture, platform, or combination.

    | Task always returns a Result object.
    | - If the task's `run` method returns Result object by itself, you'll receive this Result object
    | - If the task's `run` method returns any other object, you'll receive a Success with returned_value encapsulated
    | - If the task's `run` method fires an exception, you'll receive a Failure with the exception encapsulated

    | When creating your own tasks, please, do not forget to call `super().__init__(*args, **kwargs)` in your implementation
        to properly initialize the base class.

    :param name: Name of the task. If not provided, a random UUID will be used.
    """

    requirements: List[str] = []
    """
    A list of commands to be executed to change environment to support this task.
    """

    previous_steps: ExecutionGraphResult = {}
    r"""
    Stores results of previous steps with task name as a key and list of results as a value.
        Several results could be stored for each task if it was executed several times.
    """

    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name or str(uuid.uuid4())
        self.requirements = copy.deepcopy(self.requirements)

    def __hash__(self) -> int:
        return hash(self.name)

    def __call__(self) -> Any:
        """
        This method is called when you call the task instance as a function. It's a shortcut for `run` method.

        :return: Result of the execution
        """
        return self.run()

    def __repr__(self) -> str:
        type_ = type(self)
        module = type_.__module__
        qualname = type_.__qualname__
        return f"<{module}.{qualname} with name {self.name}>"

    def __str__(self) -> str:
        return self.__repr__()

    def add_requirement(self, command: str) -> Task:
        """
        | This method adds a requirement to the requirements of the instance of the task.
        | Please, note that the requirement would be added to the instance, not to all instances of the class.
        | Use it to provide additional requirements that should be executed only once despite the number of instances.

        :param command: Command to be executed to change environment to support this task.
        :return: self
        """
        self.requirements = copy.deepcopy(self.requirements)
        self.requirements.append(command)
        return self

    @abstractmethod
    def run(self) -> Any:
        """
        | ## This method is to be overridden by your implementation. ##
        | This is the entrypoint for the task.
        | This method should never have any arguments except `self`. Any arguments that task would use for execution
            should be provided to the constructor and used later by this method.
        | This method will always return a Result object. If this method doesn't return a Result object,
            it will be encapsulated into a Result object.

        :return: Result of the execution
        """
        raise NotImplementedError


class TaskDispatcher(ABC):
    """
    | This class is a wrapper for several tasks that are designed to implement the same functionality
        but depend on node attributes. Most often you either want to use a specific
        implementation for a specific architecture (e.g., different Tasks for Windows and Linux),
        or instantiate a task with some specific parameters for a specific node (e.g., node-specific IP address).
    | You should implement your own TaskDispatcher class and override the dispatch method.
    |
    | You also should provide and use any variables (state) that `run` method would use in the constructor of this class
        and pass them to the constructor of the task implementation.
    |
    | Dispatching is done by calling the dispatch method that you should implement.

    :param name: Name of the task. If not provided, a random UUID will be used.
    """

    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name or str(uuid.uuid4())
        """
        Name of the task.
        """

    @abstractmethod
    def dispatch(self, node: Node) -> Task:
        """
        | This method takes a node and should return and instance of the task that is designed to be executed on this node.
        | The instance could depend on the node information (such as architecture, platform, properties, etc).
        |
        | Do not forget to pass all arguments and variables to the constructor of the task implementation.

        :param node: Node instance
        :return: Task instance selected based on the node information
        """
        raise NotImplementedError


TaskElement = Union[Task, TaskDispatcher]



"""
Architecture information about nodes.
"""

from enum import Enum


class Architecture(Enum):
    """
    Enumerate possible node architectures.
    """

    LINUX_AMD64 = "linux/amd64"
    """
    AMD64-based Linux node.
    """

    LINUX_ARM64 = "linux/arm64"
    """
    ARM64-based Linux node.
    """

    WINDOWS_AMD64 = "win/amd64"
    """
    AMD64-based Windows node.
    """

    WINDOWS_ARM64 = "win/arm64"
    """
    ARM64-based Windows node.
    """

    UNKNOWN = "unknown"
    """
    Unknown architecture and operating system.
    """



"""
Experiment-related entities and classes.
"""

from __future__ import annotations

import base64
import copy
import warnings
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Union, cast

from returns.pipeline import is_successful
from returns.result import Result

from .deployment import Deployment
from .execution_graph import ExecutionGraph
from .nodes import Node, Nodes
from .types import (
    DeploymentExecutionResultRepresentation,
    ExecutionGraphResult,
    ExperimentExecutionInformationRepresentation,
    ExperimentRepresentation,
)
from .utils import LogType


class ExperimentStatus(Enum):
    """
    Represents a status of an experiment.
    """

    UNKNOWN = 0
    """
    Unknown status.
    """

    PREPARING = 1
    """
    Experiment is under preparation.
    """

    READY = 2
    """
    Experiment is prepared and ready to be started.
    """

    RUNNING = 3
    """
    Experiment is running.
    """

    FINISHED = 4
    """
    Experiment is finished.
    """

    def __json__(self) -> int:
        return self.value

    @classmethod
    def from_json(cls, value: int) -> ExperimentStatus:
        """
        Converts a JSON representation of an experiment status to an instance of ExperimentStatus.

        :param value: experiment status value.
        :return: an instance of ExperimentStatus.
        """
        return cls(value)


class Experiment:
    """
    Represents an experiment that contains a mapping of execution graphs to nodes.

    :param deployment_context: deployment context to be used by connectors.
    """

    def __init__(
        self, deployment_context: Optional[Dict[str, Dict[str, str]]] = None
    ) -> None:
        self.deployment_map: List[Deployment] = []
        """
        a list of deployments
        """

        self.deployment_context: Optional[Dict[str, Dict[str, str]]] = (
            deployment_context
        )
        """
        A dictionary that contains a context for deployments.
            Context is to be provided by connectors. 
            Format: {connector_name: {key: value}}
        """

    def append(self, node: Node, pipeline: ExecutionGraph) -> Experiment:
        """
        Append a new deployment (mapping of an execution graph to a node) to the experiment.

        :param node: a node to deploy the execution graph to.
        :param pipeline: a pipeline (execution graph) to deploy.
        :return: self.
        """

        self.deployment_map.append(Deployment(node, pipeline))
        return self

    def map(self, pipeline: ExecutionGraph, nodes: Sequence[Node]) -> Experiment:
        """
        Map an execution graph to a sequence of nodes.

        :param pipeline: a pipeline (execution graph) to deploy.
        :param nodes: a sequence of nodes to deploy the execution graph to.
        :return: self.
        """

        if isinstance(nodes, Nodes):
            raise TypeError("Expected sequence of nodes, got Nodes instead")

        for node in nodes:
            if not isinstance(node, Node):
                raise TypeError(f"Expected sequence of nodes, got {type(node)} instead")
            self.append(node, pipeline)
        return self

    def __json__(self) -> ExperimentRepresentation:
        return {
            "deployment_map": [x.__json__() for x in self.deployment_map],
            "deployment_context": self.deployment_context,
        }

    @classmethod
    def from_json(cls, data: ExperimentRepresentation) -> Experiment:
        """
        Creates an instance of Experiment from a JSON representation.

        :param data: a JSON representation of an experiment.
        :return: an instance of Experiment.
        """
        instance = cls.__new__(cls)
        instance.deployment_map = [
            Deployment.from_json(x) for x in data["deployment_map"]
        ]
        instance.deployment_context = data.get("deployment_context")
        return instance

    def __getitem__(self, item: int) -> Deployment:
        """
        Returns a deployment by index.

        :param item: an index of a deployment.
        :return: a deployment.
        """
        return self.deployment_map[item]

    def __iter__(self) -> Iterator[Deployment]:
        """
        Returns an iterator over deployments.

        :return: an iterator over deployments.
        """
        return iter(self.deployment_map)

    def __len__(self) -> int:
        """
        Returns a number of deployments in the experiment.

        :return: a number of deployments in the experiment.
        :meta public:
        """
        return len(self.deployment_map)

    def __str__(self) -> str:
        return "\n".join([f" - {x}" for x in self.deployment_map])

    def __repr__(self) -> str:
        return self.__str__()

    def __add__(self, other: Experiment) -> Experiment:
        """
        Concatenates two experiments resulting in a union of deployments.

        :param other: an experiment to concatenate with.
        :return: a new experiment.
        """

        new_map = copy.deepcopy(self)
        new_map.deployment_map.extend(other.deployment_map)
        if other.deployment_context:
            if new_map.deployment_context is None:
                new_map.deployment_context = {}
            new_map.deployment_context.update(other.deployment_context)
        return new_map


class DeploymentExecutionResult:
    """
    Stores a result (or ongoing information) of a deployment execution.

    :param node: a node that was used for deployment.
    :param serialized_execution_graph: a serialized execution graph.
    :param result: a result of a deployment execution.
    :param error: an error message if deployment failed.
    """

    def __init__(
        self,
        node: Node,
        serialized_execution_graph: bytes,
        result: Optional[bytes],
        error: Optional[str] = None,
    ):
        self.node: Node = node
        """
        A node that was used for deployment
        """

        self._execution_graph: bytes = serialized_execution_graph
        """
        A serialized exeuction graph
        """

        self._result: Optional[bytes] = result
        """
        Deployment execution result
        """

        self.error: Optional[str] = error
        """
        An error message if deployment failed.
        """

    @property
    def pipeline(self) -> ExecutionGraph:
        """
        Returns a pipeline (execution graph) that was used for the deployment.

        :return: an execution graph that was used for the deployment.
        """
        import cloudpickle

        if not self._execution_graph:
            raise ValueError("Execution graph information is not available.")

        return cast(ExecutionGraph, cloudpickle.loads(self._execution_graph))

    @property
    def result(
        self,
    ) -> Optional[Tuple[Result[ExecutionGraphResult, ExecutionGraphResult], LogType]]:
        """
        Returns a result of a deployment execution and logs.

        :return: a tuple of (execution result, logs).
        """
        import cloudpickle

        return cloudpickle.loads(self._result) if self._result else None

    def __str__(self) -> str:
        text = f"DeploymentExecutionResult:\n  Node: {self.node}\n"
        result = self.result
        if result:
            text += f"  Result: {type(result[0])}\n"
            if not isinstance(result[0], Result) or not is_successful(result[0]):
                text += f"   {result[0]}\n"
            else:
                for task_id, task_result in result[0].unwrap().items():
                    text += f"    {task_id}: {task_result}\n"

            if result[1]:
                text += f"  Logs:\n"
                for line in result[1]:
                    text += f"    {line}"
        if self.error:
            text += f"  Error: {self.error}\n"
        text += "\n"
        return text

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> DeploymentExecutionResultRepresentation:
        return {
            "node": self.node.__json__(),
            "execution_graph": base64.b64encode(self._execution_graph).decode("utf-8"),
            "result": (
                base64.b64encode(self._result).decode("utf-8") if self._result else None
            ),
            "error": self.error,
        }

    @classmethod
    def from_json(
        cls, data: DeploymentExecutionResultRepresentation
    ) -> DeploymentExecutionResult:
        """
        Returns an instance of DeploymentExecutionResult from a JSON representation.

        :param data: a JSON representation of a deployment execution result.
        :return: an instance of DeploymentExecutionResult.
        """
        try:
            execution_graph = base64.b64decode(data["execution_graph"])
        except KeyError:
            if "pipeline" in data:
                warnings.warn(
                    "The deployment was created with an older version of netunicorn. "
                    "Execution graph information would not be available."
                )
                execution_graph = b""
            else:
                raise

        return cls(
            Node.from_json(data["node"]),
            execution_graph,
            base64.b64decode(data["result"]) if data["result"] else None,
            data["error"],
        )


@dataclass(frozen=True)
class ExperimentExecutionInformation:
    """
    Stores information about an experiment execution.

    :param status: a status of an experiment execution.
    :param experiment: a definition of an experiment.
    :param execution_result: a result of an experiment execution.
    """

    status: ExperimentStatus
    """
    Describes the status of an experiment execution.
    """

    experiment: Optional[Experiment]
    """
    Stores the definition of the experiment.
    """

    execution_result: Union[None, Exception, List[DeploymentExecutionResult]]
    """
    Stores either an Exception if experiment execution failed or a list of deployment execution results.
    """

    def __str__(self) -> str:
        text = (
            f"ExperimentExecutionInformation:\n"
            f"status: {self.status}\n"
            f"experiment: \n"
            f"{self.experiment}\n"
            f"execution_result:\n"
            f"{self.execution_result}\n"
        )
        return text

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> ExperimentExecutionInformationRepresentation:
        execution_result: Union[
            None, str, List[DeploymentExecutionResultRepresentation]
        ] = None
        if isinstance(self.execution_result, list):
            execution_result = [x.__json__() for x in self.execution_result]
        elif isinstance(self.execution_result, Exception):
            execution_result = str(self.execution_result.__reduce__())
        return {
            "status": self.status.__json__(),
            "experiment": self.experiment.__json__() if self.experiment else None,
            "execution_result": execution_result,
        }

    @classmethod
    def from_json(
        cls, data: ExperimentExecutionInformationRepresentation
    ) -> ExperimentExecutionInformation:
        """
        Returns an instance of ExperimentExecutionInformation from a JSON representation.

        :param data: a JSON representation of an experiment execution information.
        :return: an instance of ExperimentExecutionInformation.
        """

        status = ExperimentStatus.from_json(data["status"])
        experiment = (
            Experiment.from_json(data["experiment"]) if data["experiment"] else None
        )
        execution_result_data = data["execution_result"]
        execution_result: Union[None, Exception, List[DeploymentExecutionResult]] = None
        if execution_result_data:
            if isinstance(execution_result_data, list):
                execution_result = [
                    DeploymentExecutionResult.from_json(x)
                    for x in execution_result_data
                ]
            else:
                execution_result = Exception(execution_result_data)
        return cls(status, experiment, execution_result)



"""
Abstrcations for Pipeline representation.
"""

from __future__ import annotations

from typing import Any, Collection, Dict, List, Optional, Union

import networkx as nx

from .environment_definitions import EnvironmentDefinition
from .execution_graph import ExecutionGraph
from .task import TaskElement

PipelineElement = Union[TaskElement, Collection[TaskElement]]


class Pipeline(ExecutionGraph):
    """
    Pipeline is a class that takes a tuple of Tasks and executes them in order.
    Each element in the tuple should be either a Task or a tuple of Tasks.
    Pipeline will execute elements in order and return the combined result of all tasks.
    If element is a tuple of tasks, these tasks would be executed in parallel.

    | The result of pipeline execution would be one of the following:
    | - Success: if all tasks succeed, then Success is returned
    | - Failure: if any task fails, then Failure is returned

    | Returning object will always contain a Result object for each task executed.
    | If early_stopping is set to True, then any task after first failed wouldn't be executed.

    :param tasks: tasks (ordered by stages) to be executed
    :param early_stopping: whether to stop executing tasks after first failure
    :param report_results: whether executor should connect core services to report pipeline results in the end
    :param environment_definition: environment definition for the pipeline
    """

    def __init__(
        self,
        tasks: Collection[PipelineElement] = (),
        early_stopping: bool = True,
        report_results: bool = True,
        environment_definition: Optional[EnvironmentDefinition] = None,
    ):
        super().__init__(
            early_stopping=early_stopping,
            report_results=report_results,
            environment_definition=environment_definition,
        )

        self.last_stage: Union[str, int] = "root"
        """
        Current last stage of the pipeline.
        """

        for element in tasks:
            self.then(element)

    @staticmethod
    def _element_to_stage(element: PipelineElement) -> List[TaskElement]:
        if not isinstance(element, Collection):
            element = [element]
        if not isinstance(element, list):
            element = list(element)
        return element

    def then(self, element: PipelineElement) -> Pipeline:
        """
        Add a task or list of tasks as a separate stage to the end of the pipeline.

        :param element: a task or tuple of tasks to be added
        :return: self
        """
        element = self._element_to_stage(element)

        initial_stage = self.last_stage
        next_stage = self.last_stage + 1 if isinstance(self.last_stage, int) else 1
        self.graph.add_edges_from([(initial_stage, x) for x in element])
        self.graph.add_edges_from([(x, next_stage) for x in element])
        self.last_stage = next_stage

        return self

    def copy(self) -> Pipeline:
        """
        Return a copy of the pipeline.

        :return: a copy of the pipeline
        """
        pipeline = Pipeline(
            early_stopping=self.early_stopping,
            report_results=self.report_results,
            environment_definition=self.environment_definition,
        )

        pipeline.graph = self.graph.copy()
        return pipeline

    def __str__(self) -> str:
        successors = nx.dfs_successors(self.graph, "root")
        stages = {
            x: y for x, y in successors.items() if isinstance(x, int) or x == "root"
        }

        return f"Pipeline({self.name}): {stages}"

    def __repr__(self) -> str:
        return str(self)


class CyclePipeline(Pipeline):
    """
    CyclePipeline is a Pipeline that will be executed several times. All defined stages would be executed
    and then the execution would continue from the first stage again.
    """

    def __init__(
        self,
        cycles: Optional[int] = None,
        tasks: Collection[PipelineElement] = (),
        early_stopping: bool = True,
        report_results: bool = True,
        environment_definition: Optional[EnvironmentDefinition] = None,
    ):
        self.edge_params: Dict[str, Any] = {"type": "weak"}
        if cycles is not None:
            assert isinstance(cycles, int)
            if cycles <= 2:
                raise ValueError(
                    f"Number of cycles should be at least 2, current value: {cycles}"
                )
            self.edge_params["counter"] = cycles - 1

        super().__init__(
            tasks=tasks,
            early_stopping=early_stopping,
            report_results=report_results,
            environment_definition=environment_definition,
        )

        self.graph.add_edge("root", "root", **self.edge_params)

    def then(self, element: PipelineElement) -> Pipeline:
        """
        Add a task or list of tasks as a separate stage to the end of the pipeline.

        :param element: a task or tuple of tasks to be added
        :return: self
        """
        element = self._element_to_stage(element)

        initial_stage = self.last_stage
        if self.graph.has_edge(initial_stage, "root"):
            self.graph.remove_edge(initial_stage, "root")

        next_stage = self.last_stage + 1 if isinstance(self.last_stage, int) else 1
        self.graph.add_edges_from([(initial_stage, x) for x in element])
        self.graph.add_edges_from([(x, next_stage) for x in element])
        self.last_stage = next_stage
        self.graph.add_edge(next_stage, "root", **self.edge_params)

        return self
